#include <Windows.h>

#include <cstdio>

int wmain(int argc, wchar_t** argv) {
  if (argc != 2) {
    wprintf(L"Usage: %s <new user SID>\n", argv[0]);
    return 1;
  }

  //
  // Open a handle to the layered GroupMembership key in the delta hive.
  //

  HKEY hGroupPolicyKey;
  LONG st = RegOpenKeyExW(HKEY_USERS,
                          L"Test\\Microsoft\\Windows\\CurrentVersion\\Group Policy",
                          0,
                          KEY_READ,
                          &hGroupPolicyKey);

  if (st != ERROR_SUCCESS) {
    wprintf(L"RegOpenKeyExW #1 failed with error %d\n", st);
    return 1;
  }

  HKEY hUserSidKey;
  st = RegOpenKeyExW(hGroupPolicyKey,
                     argv[1],
                     0,
                     KEY_READ,
                     &hUserSidKey);

  if (st != ERROR_SUCCESS) {
    wprintf(L"RegOpenKeyExW #2 failed with error %d\n", st);
    return 1;
  }

  HKEY hGroupMembershipKey;
  st = RegOpenKeyExW(hUserSidKey,
                     L"GroupMembership",
                     0,
                     KEY_READ,
                     &hGroupMembershipKey);

  if (st != ERROR_SUCCESS) {
    wprintf(L"RegOpenKeyExW #3 failed with error %d\n", st);
    return 1;
  }

  wprintf(L"Successfully opened the key... ");
  getchar();

  //
  // Enumerate the key and print out the value names.
  // 

  wprintf(L"Key values:\n");

  for (DWORD i = 0;; i++) {
    WCHAR ValueName[100];
    DWORD cchValueName = sizeof(ValueName) / sizeof(ValueName[0]);

    st = RegEnumValueW(hGroupMembershipKey,
                       i,
                       ValueName,
                       &cchValueName,
                       NULL,
                       NULL,
                       NULL,
                       NULL);

    if (st == ERROR_NO_MORE_ITEMS) {
      break;
    } else if (st != ERROR_SUCCESS) {
      printf("RegEnumValueA failed with error %d\n", st);
      return 1;
    }

    wprintf(L"%d. %s\n", i + 1, ValueName);
  }

  //
  // Close handles etc.
  //

  RegCloseKey(hGroupMembershipKey);
  RegCloseKey(hUserSidKey);
  RegCloseKey(hGroupPolicyKey);

  return 0;
}
